import pytest
from agepy import ageplot

def test_use_with_age_styles():
    for style in ageplot.age_styles:
        ageplot.use(style)

def test_use_with_mpl_styles():
    for style in ageplot.mpl_styles:
        ageplot.use(style)

def test_use_with_invalid_style():
    with pytest.raises(ValueError):
        ageplot.use("non_existent_style")

def test_use_with_multiple_styles():
    ageplot.use(["age", "pccp"])

def test_use_with_invalid_input():
    with pytest.raises(TypeError):
        ageplot.use(123)

def test_figsize_with_media():
    for medium in ageplot.figsize.media:
        test_figsize = ageplot.figsize(medium)
        assert test_figsize.h <= test_figsize.hmax

def test_figsize_with_invalid_medium():
    with pytest.raises(ValueError):
        ageplot.figsize("non_existent_medium")



"""Plotting module containing custom AGE matplotlib styles and functions
for creating nice plots.

Attributes
----------
age_styles: list
    List of strings containing all the available custom AGE style
    sheets.
mpl_styles: list
    List of strings containing all the available matplotlib style
    sheets.
colors: list
    List of AGE colors (taken from the seaborn colorblind palette).

"""

import matplotlib.pyplot as plt

age_styles = ["age", "tex", "nature", "prl", "pccp", "powerpoint",
              "latexbeamer"]
mpl_styles = ["default"]
mpl_styles.extend(plt.style.available)

colors = [
    "#0173b2", "#de8f05", "#029e73", "#d55e00", "#cc78bc", "#ca9161",
    "#fbafe4", "#949494", "#ece133", "#56b4e9"
]


def use(styles):
    """Function calling :py:func:`plt.style.use` for easier access to
    the custom AGE matplotlib style sheets.

    Notes
    -----
    All style rcParams are reset to the matplotlib default before
    loading the specified styles.

    Warnings
    --------
    Compatibility between styles is not guaranteed.

    Parameters
    ----------
    styles: str or list of string
        Styles to be loaded using :py:func:`plt.style.use`. Available styles
        can be viewed by calling :py:data:`ageplot.age_styles` and
        :py:data:`ageplot.mpl_styles`.

    Examples
    --------
    How to use the AGE style in conjunction with a style setting the
    figuresize, fontsizes and linewidths:

    >>> import matplotlib.pyplot as plt
    >>> from agepy import ageplot
    >>> ageplot.use(["age", "pccp"])
    >>> plt.plot([1, 2, 3], [4, 5, 6])

    """
    load_styles = []

    # Check if styles are available
    if isinstance(styles, list):
        for style in styles:
            if style in age_styles:
                load_styles.append("agepy.ageplot." + style)
            elif style in mpl_styles:
                load_styles.append(style)
            else:
                raise ValueError(style + " is not an available style.")
    elif isinstance(styles, str):
        if styles in age_styles:
            load_styles.append("agepy.ageplot." + styles)
        elif styles in mpl_styles:
            load_styles.append(styles)
        else:
            raise ValueError(styles + " is not an available style.")
    else:
        raise TypeError("Expected str or list of strings specifying styles.")

    plt.style.use("default")  # reset rcParams before applying the style
    plt.style.use(load_styles)  # apply the selected styles


class figsize():
    """Class for choosing the appropriate size of matplotlib figures.

    This class provides access to the width and height of the available
    space in different media in order to choose a figure size for
    matplotlib plots.

    Parameters
    ----------
    medium: str, optional
        Name of the medium. Media, for which the size is
        implemented, can be viewed with :py:attr:`figsize.media`.
        Default: None

    Attributes
    ----------
    w: float
        Recommended width for a figure. Most of the time this will be
        equivalent to the full available width.
    h: float
        Recommended height corresponding to the width
        (width * 3 / 4).
    wh: tuple
        Tuple (w, h) containing the width and recommended height.
        If there is a corresponding style sheet, the default figure
        size will be set to (w, h) by :py:func:`agepy.ageplot.use`.
    hmax: float
        Height in inches available for a figure.
    media: list
        Class wide attribute containing a list of media, for which
        the figsize is implemented.

    Examples
    --------
    A simple example for the PCCP journal:

    >>> from agepy import ageplot
    >>> my_figsize = ageplot.figsize("pccp")
    >>> my_figsize.wh
    (3.54, 2.6550000000000002)

    This could be used for a matplotlib plot like this:

    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots(1, figsize=(my_figsize.w, my_figsize.hmax / 3))
    >>> plt.plot([1, 2, 3], [4, 5, 6])

    """

    # (textwidth, textheight)
    # If the journal uses multiple columns, columnwidth instead of
    # textwidth is used.
    _pagesizes = {
        "pccp": (3.54, 9.54),
        "powerpoint": (5, 5.625),  # 0.5 of full width
        "latexbeamer": (2.766, 3.264),  # 0.5 of full width
        "nature2col": (180 / 25.4, 225 / 25.4),
        "nature1col": (88 / 25.4, 220 / 25.4),
        "prl1col": (3.375, 8.7),
        "prl2col": (3.375 * 1.5, 8.7),
    }
    media = list(_pagesizes.keys())

    def __init__(self, medium=None):
        if medium is None:
            self.w = 6.4
            self.h = 4.8
            self.hmax = None
        elif medium in self.media:
            self.w = self._pagesizes[medium][0]
            self.h = self._pagesizes[medium][0] * 0.75
            self.hmax = self._pagesizes[medium][1]
        else:
            raise ValueError(medium + " is not an available medium.")

        self.wh = (self.w, self.h)



import numpy as np
import matplotlib.pyplot as plt
from agepy.spec import coincidence

def test_plot_coinc_map():
    # AI generated test. Think about more meaningful tests.
    # Test data
    coincmap = np.random.rand(10, 10)
    xedges = np.linspace(0, 1, 11)
    yedges = np.linspace(0, 1, 11)
    figsize = (8, 6)
    cmap = 'YlOrRd'
    title = "Coincidence Map"
    norm = None
    vmin = 1
    vmax = None
    num = 1
    xlabel = "early electron kinetic energy"
    ylabel = "late electron kinetic energy"

    # Call the function
    fig, (ax_coinc, ax_x, ax_y) = coincidence.plot_coinc_map(
        coincmap, xedges, yedges, figsize=figsize, cmap=cmap, title=title,
        norm=norm, vmin=vmin, vmax=vmax, num=num, xlabel=xlabel, ylabel=ylabel
    )

    # Assertions
    assert isinstance(fig, plt.Figure)
    assert isinstance(ax_coinc, plt.Axes)
    assert isinstance(ax_x, plt.Axes)
    assert isinstance(ax_y, plt.Axes)

    # Cleanup
    plt.close(fig)


from __future__ import annotations
from typing import Union, Tuple, Sequence
import numpy as np


# Classes implementing the various types of anodes

class PositionAnode(object):
    """Generic anode as a parent class for all anodes.

    Parameters
    ----------
    angle: int or float
        Rotation of detector.
    scale: tuple
        scale = (scale_x, scale_y)
    offset: tuple
        offset = (offset_x, offset_y)

    """

    def __init__(self, angle: Union[int, float], scale: Tuple[float, float],
                 offset: Tuple[float, float]) -> None:
        self.transform = np.zeros((2, 2), dtype=float)
        self.offset = np.zeros((2,), dtype=float)
        pre_v, post_v = angle * np.pi/180, 0 * np.pi/180
        pre_sin, post_sin = np.sin(pre_v), np.sin(post_v)
        pre_cos, post_cos = np.cos(pre_v), np.cos(post_v)
        # First, rotate by the pre matrix, then scale and
        # afterwards rotate again by the post matrix.
        # The components are the product of these three matrices in
        # order.
        self.transform[0,0] = (pre_cos * post_cos * (+scale[0])
                               + pre_sin * post_sin * (-scale[1]))
        self.transform[1,0] = (pre_sin * post_cos * (+scale[0])
                               + pre_cos * post_sin * (+scale[1]))
        self.transform[0,1] = (pre_cos * post_sin * (-scale[0])
                               + pre_sin * post_cos * (-scale[1]))
        self.transform[1,1] = (pre_sin * post_sin * (-scale[0])
                               + pre_cos * post_cos * (+scale[1]))
        # +0.5 for the offset introduced by centering the image on
        # the origin.
        self.offset[0] = offset[0] + 0.5
        self.offset[1] = offset[1] + 0.5        

    def process(self, rows):
        """Needs documentation.

        """
        rows = rows[:, :2] - 0.5
        rows = rows.dot(self.transform) + self.offset

        return rows


class WsaAnode(PositionAnode):
    """Needs documentation.

    """

    def __init__(self, angle, scale=(3, 3), offset=(0.8, 0.8),
                 scale_a=0.8, scale_b=0.65, scale_c=1.5,
                 offset_a=0, offset_b=0, offset_c=0):
        super().__init__(angle, scale, offset)

        self.scale_a = scale_a
        self.scale_b = scale_b
        self.scale_c = scale_c

        self.offset_a = offset_a
        self.offset_b = offset_b
        self.offset_c = offset_c

    def process(self, rows):
        a = self.scale_a * rows[:, 0] + self.offset_a
        b = self.scale_b * rows[:, 1] + self.offset_b
        abc = a + b + self.scale_c * rows[:, 2] + self.offset_c

        pos = np.empty((a.shape[0], 2), dtype=float)
        pos[:, 0] = a / abc
        pos[:, 1] = b / abc

        return super().process(pos)


class PocoAnode(PositionAnode):
    """Needs documentation.

    """

    def __init__(self, angle):
        super().__init__(angle, (0.0005, 0.0005), (0, 0))


class DldAnode(PositionAnode):
    """DLD anode parent class.

    """

    def __init__(self, angle):
        super().__init__(angle, (1e-05, 1e-05), (0, 0))


class DldAnodeXY(DldAnode):
    """DLD anode for the xy mode in metro and a Roentdek OpenFace
    quad delay-line detector.

    Parameters
    ----------
    angle: int or float
        Rotation of detector.

    """

    def process(self, rows: np.ndarray) -> np.ndarray:
        """Processes the raw data probably found in dld_rd#raw.

        Parameters
        ----------
        rows: np.ndarray
            Raw data from the detector.

        Returns
        -------
        np.ndarray
            Processed data in form of xy values.
        
        """
        x = rows[:, 0] - rows[:, 1]
        y = rows[:, 2] - rows[:, 3]

        # TODO optimize calibration, the factor 1.06 is just by eye for now;
        # also, a cross correction should be done (software or hardware)
        ratio_xy = 1.06

        pos = np.empty((x.shape[0], 2), dtype=float)
        pos[:, 0] = x
        pos[:, 1] = y * ratio_xy

        return super().process(pos)


class DldAnodeUVW(DldAnode):
    """Anode used with the uvw mode in metro and a Roentdek Hex
    delay-line detector.

    Parameters
    ----------
    angle: int or float
        Rotation of detector.

    """

    def process(self, rows: np.ndarray) -> np.ndarray:
        """Processes the raw data probably found in dld_rd#raw.

        Parameters
        ----------
        rows: np.ndarray
            Raw data from the detector.

        Returns
        -------
        np.ndarray
            Processed data in form of xy values.
        
        """
        u = rows[:, 0] - rows[:, 1]
        v = rows[:, 2] - rows[:, 3]
        w = rows[:, 4] - rows[:, 5]

        pos = np.empty((u.shape[0], 2), dtype=float)
        pos[:, 0] = (2*u - v - w) / 3
        pos[:, 1] = (v - w) / np.sqrt(3)

        return super().process(pos)


class DldAnodeUV(DldAnode):
    """Needs documentation.

    """

    def process(self, rows):
        u = rows[:, 0] - rows[:, 1]
        v = rows[:, 2] - rows[:, 3]

        pos = np.empty((u.shape[0], 2), dtype=float)
        pos[:, 0] = u
        pos[:, 1] = (u + 2*v) / 1.7321

        return super().process(pos)


class DldAnodeUW(DldAnode):
    """Needs documentation.

    """

    def process(self, rows):
        u = rows[:, 0] - rows[:, 1]
        w = rows[:, 2] - rows[:, 3]

        pos = np.empty((u.shape[0], 2), dtype=float)
        pos[:, 0] = u
        pos[:, 1] = (u + 2*w) / -1.7321

        return super().process(pos)


class DldAnodeVW(DldAnode):
    """Needs documentation.

    """

    def process(self, rows):
        v = rows[:, 0] - rows[:, 1]
        w = rows[:, 2] - rows[:, 3]

        pos = np.empty((v.shape[0], 2), dtype=float)
        pos[:, 0] = - v - w
        pos[:, 1] = (v - w) / 1.7321

        return super().process(pos)


class Old_PositionAnode(PositionAnode):
    """Needs documentation.

    """

    def __init__(self, angle, scale, offset):
        self.transform = np.zeros((2,2), dtype=float)
        self.offset = np.zeros((2,), dtype=float)

        self.offset[0] = offset[0] + 0.5/scale[0]
        self.offset[1] = offset[1] + 0.5/scale[1]

        v = angle * 3.14/180
        sin, cos = np.sin(v), np.cos(v)

        self.transform[0,0] = cos * scale[0]
        self.transform[1,0] = sin * scale[0]
        self.transform[0,1] = -sin * scale[1]
        self.transform[1,1] = cos * scale[1]

    def process(self, rows):
        rows = rows - self.offset
        rows = rows.dot(self.transform) + 0.5

        return rows


class Old_WsaAnode(Old_PositionAnode):
    """Needs documentation.

    """

    def __init__(self, angle, scale=(2.65, 2.70), offset=(0.04, 0.06),
                 scale_a=0.80, scale_b=0.65, scale_c=1.50,
                 offset_a=0, offset_b=0, offset_c=0):
        super().__init__(angle, scale, offset)

        self.scale_a = scale_a
        self.scale_b = scale_b
        self.scale_c = scale_c

        self.offset_a = offset_a
        self.offset_b = offset_b
        self.offset_c = offset_c

    def process(self, rows):
        a = self.scale_a * rows[:, 0] + self.offset_a
        b = self.scale_b * rows[:, 1] + self.offset_b
        abc = a + b + self.scale_c * rows[:, 2] + self.offset_c

        pos = np.empty((a.shape[0], 2), dtype=float)
        pos[:, 0] = a / abc
        pos[:, 1] = b / abc

        return super().process(pos)


class Old_DldAnode(Old_PositionAnode):
    """Needs documentation.

    """

    def __init__(self, angle):
        super().__init__(angle, (0.0005, 0.0005), (-1024, -1024))

    def process(self, rows):
        u = rows[:, 0] - rows[:, 1]
        v = rows[:, 2] - rows[:, 3]
        w = rows[:, 4] - rows[:, 5]

        pos = np.empty((u.shape[0], 2), dtype=float)
        pos[:, 0] = (2*u - v - w) / 3
        pos[:, 1] = (v - w) / 1.7321

        return super().process(pos)
